from enum import Enum


class Endpoint(str, Enum):
    BASE = "base"
    CREATE = "create"
    READ = "read"
    READ_ONE = "read_one"
    UPDATE = "update"
    DELETE = "delete"


class Field(str, Enum):
    PRIMARY = "primary"


class Method(str, Enum):
    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    PUT = "PUT"
    DELETE = "DELETE"



from resty.exceptions import (
    NotFoundError,
    BadRequestError,
    UnauthorizedError,
    MethodNotAllowedError,
    InternalServerError,
    ForbiddenError,
)

STATUS_ERRORS = {
    400: BadRequestError,
    401: UnauthorizedError,
    403: ForbiddenError,
    404: NotFoundError,
    405: MethodNotAllowedError,
    500: InternalServerError,
}



from typing import Container, Mapping

from resty.middlewares import BaseResponseMiddleware
from resty.types import Response
from resty.constants import STATUS_ERRORS
from resty.exceptions import HTTPError


class StatusCheckingMiddleware(BaseResponseMiddleware):
    def __init__(
            self,
            errors: Mapping[int, type[Exception]] = None,
            default_error: type[Exception] = HTTPError,
    ):
        self._errors = errors or STATUS_ERRORS
        self._default_error = default_error

    @staticmethod
    def _check_status(actual: int, expected: int | Container[int] = 200) -> bool:
        if isinstance(expected, Container):
            return actual in expected
        return actual == expected

    def _raise_error(self, status: int, *args):
        exc = self._errors.get(status, self._default_error)

        try:
            raise exc(*args)
        except TypeError:
            raise exc()

    async def __call__(self, response: Response, **kwargs):
        actual_status = response.status
        expected_status = kwargs.pop(
            "expected_status",
            {200, 201},
        )

        check_status = kwargs.pop("check_status", True)

        if check_status and not self._check_status(actual_status, expected_status):
            self._raise_error(actual_status, response)



from contextlib import contextmanager
from typing import Iterable

from resty.middlewares.types import BaseMiddlewareManager, BaseMiddleware


class MiddlewareManager(BaseMiddlewareManager):
    def __init__(self, middlewares: Iterable = None):
        self._middlewares = []

        self.add_middlewares(*middlewares or ())

    @property
    def middlewares(self) -> Iterable[BaseMiddleware]:
        return tuple(self._middlewares)

    def add_middleware(self, middleware: BaseMiddleware):
        if not isinstance(middleware, BaseMiddleware):
            raise TypeError("Middleware must inherit the base type BaseMiddleware")

        if middleware not in self._middlewares:
            self._middlewares.append(middleware)

    def add_middlewares(self, *middlewares: BaseMiddleware):
        for middleware in middlewares:
            self.add_middleware(middleware)

    def remove_middleware(self, middleware: BaseMiddleware):
        if middleware in self._middlewares:
            self._middlewares.remove(middleware)

    def remove_middlewares(self, *middlewares: BaseMiddleware):
        for middleware in middlewares:
            self.remove_middleware(middleware)

    @contextmanager
    def middleware(self, *middlewares: BaseMiddleware):
        self.add_middlewares(*middlewares)
        yield
        self.remove_middlewares(*middlewares)

    async def __call__(
        self, *args, base: type[BaseMiddleware] = BaseMiddleware, **kwargs
    ):
        for middleware in self._middlewares:
            if isinstance(middleware, base):
                await middleware(*args, **kwargs)



from abc import ABC, abstractmethod
from typing import Mapping, Iterable

from resty.types import Schema


class BaseSerializer(ABC):
    @classmethod
    @abstractmethod
    def serialize(
            cls,
            obj: Schema,
            **kwargs,
    ) -> Mapping: ...

    @classmethod
    @abstractmethod
    def serialize_many(
            cls,
            objs: Iterable[Schema],
            **kwargs,
    ) -> Iterable: ...

    @classmethod
    @abstractmethod
    def deserialize[T: Schema](
            cls,
            schema: type[T],
            data: Mapping,
            **kwargs,
    ) -> T: ...

    @classmethod
    @abstractmethod
    def deserialize_many[T: Schema](
            cls,
            schema: type[T],
            data: Iterable,
            **kwargs,
    ) -> Iterable[T]: ...



import json

import httpx

from resty.clients.types import (
    BaseRESTClient,
    Request,
    Response,
)
from resty.middlewares import (
    MiddlewareManager,
    BaseRequestMiddleware,
    BaseResponseMiddleware,
    StatusCheckingMiddleware,
    BaseMiddlewareManager,
)
from resty.exceptions import ConnectError


class RESTClient(BaseRESTClient):

    def __init__(
        self,
        httpx_client: httpx.AsyncClient = None,
        check_status: bool = True,
        middleware_manager: BaseMiddlewareManager = None,
    ):
        self.middlewares = middleware_manager or MiddlewareManager()
        self._xclient = httpx_client or httpx.AsyncClient()

        if check_status:
            self.middlewares.add_middleware(StatusCheckingMiddleware())

    async def _make_xrequest(self, request: Request) -> httpx.Response:
        try:
            return await self._xclient.request(
                method=request.method.value,
                url=request.url,
                headers=request.headers,
                json=request.json,
                data=request.data,
                params=request.params,
                cookies=request.cookies,
                follow_redirects=request.redirects,
                timeout=request.timeout,
            )
        except httpx.ConnectError:
            raise ConnectError(url=request.url)

    @staticmethod
    def _extract_json_data(xresponse: httpx.Response) -> dict | list:
        try:
            data = xresponse.json()
        except json.decoder.JSONDecodeError:
            data = {}

        return data

    async def _parse_xresponse(
        self, request: Request, xresponse: httpx.Response
    ) -> Response:
        return Response(
            request=request,
            status=xresponse.status_code,
            json=self._extract_json_data(xresponse=xresponse),
            content=xresponse.content,
            text=xresponse.text,
            middleware_options=request.middleware_options,
        )

    async def _make_request(self, request: Request) -> Response:
        xresponse = await self._make_xrequest(request=request)
        response = await self._parse_xresponse(request=request, xresponse=xresponse)
        return response

    async def _call_middlewares(self, reqresp: Request | Response):
        await self.middlewares(
            reqresp,
            base=(
                BaseRequestMiddleware
                if isinstance(reqresp, Request)
                else BaseResponseMiddleware
            ),
            **reqresp.middleware_options
        )

    async def request(self, request: Request) -> Response:
        await self._call_middlewares(reqresp=request)

        response = await self._make_request(request)

        await self._call_middlewares(reqresp=response)

        return response



from functools import cache
from urllib.parse import urljoin, urlparse

from resty.enums import Endpoint
from resty.exceptions import URLBuildingError
from resty.managers.types import BaseURLBuilder, Endpoints


class URLBuilder(BaseURLBuilder):
    @classmethod
    def _get_endpoint_url(cls, endpoints: Endpoints, endpoint: Endpoint) -> str | None:
        url = endpoints.get(endpoint, endpoints.get(Endpoint.BASE, None))
        return url

    @classmethod
    def _inject_params(cls, url: str, **params) -> str:
        try:
            return url.format(**params)
        except KeyError as e:
            raise URLBuildingError(f"Missing '{e.args[0]}' in {url}")

    @classmethod
    @cache
    def _normalize_url(cls, url: str | None) -> str:
        if not url:
            return ""

        if not url.endswith('/'):
            return url + '/'
        return url

    @classmethod
    def build(
            cls, endpoints: Endpoints, endpoint: Endpoint, base_url: str = None, **kwargs
    ) -> str:

        endpoint_url = cls._get_endpoint_url(endpoints=endpoints, endpoint=endpoint)

        if endpoint_url:
            url = urljoin(
                cls._normalize_url(url=base_url),
                endpoint_url
            )
        else:
            url = base_url or ""

        return cls._inject_params(url=url, **kwargs)



from resty.middlewares.types import (
    BaseMiddleware,
    BaseMiddlewareManager,
    BaseResponseMiddleware,
    BaseRequestMiddleware,
)
from resty.middlewares.managers import MiddlewareManager
from resty.middlewares.status import StatusCheckingMiddleware

__all__ = [
    "MiddlewareManager",
    "BaseMiddleware",
    "BaseMiddlewareManager",
    "BaseResponseMiddleware",
    "BaseRequestMiddleware",
    "StatusCheckingMiddleware",
]



from dataclasses import dataclass, field
from typing import Mapping, Iterable

from pydantic import BaseModel

from resty.enums import Method, Endpoint

Schema = BaseModel


@dataclass
class Request:
    url: str
    method: Method
    endpoint: Endpoint = None
    data: Mapping | Iterable = None
    json: Mapping | Iterable = None
    timeout: int | None = None
    params: dict = field(default_factory=dict)
    headers: dict = field(default_factory=dict)
    cookies: dict = field(default_factory=dict)
    redirects: bool = False
    middleware_options: dict = field(default_factory=dict)


@dataclass
class Response:
    request: Request
    status: int
    content: bytes = None
    text: str = None
    json: list | dict = None
    middleware_options: dict = field(default_factory=dict)



from typing import Iterable
from contextlib import contextmanager
from abc import ABC, abstractmethod

from resty.types import Request, Response


class BaseMiddleware(ABC):
    @abstractmethod
    def __call__(self, *args, **kwargs): ...


class BaseRequestMiddleware(BaseMiddleware, ABC):
    @abstractmethod
    def __call__(self, request: Request, **kwargs): ...


class BaseResponseMiddleware(BaseMiddleware, ABC):
    @abstractmethod
    def __call__(self, response: Response, **kwargs): ...


class BaseMiddlewareManager(ABC):
    @property
    @abstractmethod
    def middlewares(self) -> Iterable[BaseMiddleware]: ...

    @abstractmethod
    def add_middleware(self, middleware: BaseMiddleware): ...

    @abstractmethod
    def add_middlewares(self, *middlewares: BaseMiddleware): ...

    @abstractmethod
    def remove_middleware(self, middleware: BaseMiddleware): ...

    @abstractmethod
    def remove_middlewares(self, *middlewares: BaseMiddleware): ...

    @contextmanager
    @abstractmethod
    def middleware(self, *middlewares: BaseMiddleware): ...

    @abstractmethod
    async def __call__(
        self, *args, base: type[BaseMiddleware] = BaseMiddleware, **kwargs
    ): ...



import inspect
from typing import Mapping, Iterable, Callable

from resty.clients import BaseRESTClient
from resty.enums import Endpoint, Method, Field
from resty.types import Schema, Response, Request
from resty.serializers import Serializer, BaseSerializer
from resty.managers.types import BaseManager, ResponseType
from resty.managers.builders import URLBuilder


class Manager(BaseManager):
    serializer_class = Serializer
    url_builder_class = URLBuilder

    @classmethod
    def get_serializer(cls, **kwargs) -> type[BaseSerializer]:
        serializer = kwargs.pop('serializer', cls.serializer_class)

        if not serializer:
            raise RuntimeError('Serializer not specified')

        if not (
                isinstance(serializer, BaseSerializer)
                or inspect.isclass(serializer)
                and issubclass(serializer, BaseSerializer)
        ):
            raise RuntimeError('The serializer must be a subclass of BaseSerializer')

        return serializer

    @classmethod
    def get_method(cls, endpoint: Endpoint, **kwargs) -> Method:
        method = cls.methods.get(endpoint)
        if not method:
            raise RuntimeError(f'Method not specified for endpoint: {endpoint}')

        return method

    @classmethod
    def get_field(cls, field: Field) -> str:
        field = cls.fields.get(field)

        if not field:
            raise RuntimeError(f'Field not specified: {field}')

        return field

    @classmethod
    def get_pk(cls, obj: Schema | Mapping) -> any:
        field = cls.get_field(Field.PRIMARY)

        if isinstance(obj, Mapping):
            return obj.get(field)

        return getattr(obj, field, None)

    @classmethod
    def _get_pk(cls, obj_or_pk: any) -> any:
        if isinstance(obj_or_pk, Mapping | Schema):
            return cls.get_pk(obj=obj_or_pk)
        return obj_or_pk

    @classmethod
    def _deserialize(cls, schema: type[Schema], data: any, **kwargs):
        serializer = cls.get_serializer(**kwargs)
        if isinstance(data, Mapping):
            return serializer.deserialize(schema=schema, data=data, **kwargs)
        return serializer.deserialize_many(schema=schema, data=data, **kwargs)

    @classmethod
    def _serialize(cls, obj: Schema, **kwargs):
        serializer = cls.get_serializer(**kwargs)
        return serializer.serialize(obj=obj, **kwargs)

    @classmethod
    async def _make_request(cls, client: BaseRESTClient, request: Request) -> Response:
        return await client.request(request=request)

    @classmethod
    def _prepare_url(cls, endpoint: Endpoint, **kwargs) -> str:
        url = kwargs.pop('url', None)
        base_url = kwargs.pop('base_url', None)

        if isinstance(url, str):
            return url

        return cls.url_builder_class.build(
            endpoints=cls.endpoints,
            endpoint=endpoint,
            base_url=base_url or cls.url,
            **kwargs
        )

    @classmethod
    def _prepare_json(cls, **kwargs):
        obj = kwargs.pop('obj', None)

        if isinstance(obj, dict | list | set | tuple):
            return obj
        elif isinstance(obj, Schema):
            return cls._serialize(obj, **kwargs)
        return {}

    @classmethod
    def _prepare_request(cls, endpoint: Endpoint, **kwargs) -> Request:
        return Request(
            url=cls._prepare_url(endpoint=endpoint, **kwargs),
            method=cls.get_method(endpoint, **kwargs),
            endpoint=endpoint,
            data=kwargs.pop('data', {}),
            json=cls._prepare_json(**kwargs),
            timeout=kwargs.pop('timeout', None),
            params=kwargs.pop('params', {}),
            headers=kwargs.pop('headers', {}),
            cookies=kwargs.pop('cookies', {}),
            redirects=kwargs.pop('redirects', False),
            middleware_options=kwargs.copy(),
        )

    @classmethod
    def _handle_response(cls, response: Response, response_type: ResponseType, **kwargs) -> any:
        if not response:
            return

        if inspect.isclass(response_type):
            if issubclass(response_type, dict | list | tuple | set):
                return response_type(response.json)
            elif issubclass(response_type, Schema):
                return cls._deserialize(schema=response_type, data=response.json, **kwargs)

        if callable(response_type):
            try:
                return response_type(response)
            except TypeError:
                pass

        return response.json

    @classmethod
    async def create[T: Schema](cls, client: BaseRESTClient, obj: Schema | Mapping, response_type: ResponseType = None,
                                **kwargs) -> T | None:
        request = cls._prepare_request(endpoint=Endpoint.CREATE, obj=obj, **kwargs)
        response = await cls._make_request(client=client, request=request)
        return cls._handle_response(response=response, response_type=response_type, **kwargs)

    @classmethod
    async def read[T: Schema](cls, client: BaseRESTClient, response_type: ResponseType = None, **kwargs) -> Iterable[T]:
        request = cls._prepare_request(endpoint=Endpoint.READ, **kwargs)
        response = await cls._make_request(client=client, request=request)
        return cls._handle_response(response=response, response_type=response_type, **kwargs)

    @classmethod
    async def read_one[T: Schema](cls, client: BaseRESTClient, obj_or_pk: Schema | Mapping | any,
                                  response_type: ResponseType = None, **kwargs) -> T:

        request = cls._prepare_request(
            endpoint=Endpoint.READ_ONE,
            pk=cls._get_pk(obj_or_pk=obj_or_pk),
            **kwargs
        )
        response = await cls._make_request(client=client, request=request)
        return cls._handle_response(response=response, response_type=response_type, **kwargs)

    @classmethod
    async def update[T: Schema](cls, client: BaseRESTClient, obj: Schema | Mapping, response_type: ResponseType = None,
                                **kwargs) -> T | None:
        request = cls._prepare_request(
            endpoint=Endpoint.UPDATE,
            pk=kwargs.pop('pk', cls.get_pk(obj)),
            obj=obj,
            **kwargs
        )
        response = await cls._make_request(client=client, request=request)
        return cls._handle_response(response=response, response_type=response_type, **kwargs)

    @classmethod
    async def delete[T: Schema](cls, client: BaseRESTClient, obj_or_pk: Schema | Mapping | any,
                                response_type: ResponseType = None,
                                **kwargs) -> T | None:
        request = cls._prepare_request(
            endpoint=Endpoint.DELETE,
            pk=cls._get_pk(obj_or_pk=obj_or_pk),
            **kwargs
        )
        response = await cls._make_request(client=client, request=request)
        return cls._handle_response(response=response, response_type=response_type, **kwargs)



__title__ = "Resty-Client"
__version__ = "0.0.5"
__description__ = """RestyClient is a simple, easy-to-use Python library for interacting with REST APIs using Pydantic's 
powerful data validation and deserialization tools."""



from abc import ABC, abstractmethod

from resty.middlewares import BaseMiddlewareManager
from resty.types import Request, Response


class BaseRESTClient(ABC):
    middlewares: BaseMiddlewareManager

    @abstractmethod
    async def request(self, request: Request) -> Response: ...



from abc import ABC, abstractmethod
from typing import Mapping, Iterable, Callable

from resty.enums import Endpoint, Field, Method
from resty.serializers import BaseSerializer
from resty.clients import BaseRESTClient
from resty.types import Schema, Response

type Endpoints = Mapping[Endpoint, str]
type Fields = Mapping[Field, str]
type Methods = Mapping[Endpoint, Method]
type ResponseType = type[Schema] | type[Mapping | Iterable] | Callable[[Response, ], any] | None


class BaseURLBuilder(ABC):
    @classmethod
    @abstractmethod
    def build(
            cls,
            base: str,
            endpoints: Endpoints,
            endpoint: Endpoint,
            **kwargs,
    ) -> str: ...


class BaseManager(ABC):
    url: str = None
    methods: Methods = {
        Endpoint.CREATE: Method.POST,
        Endpoint.READ: Method.GET,
        Endpoint.READ_ONE: Method.GET,
        Endpoint.UPDATE: Method.PATCH,
        Endpoint.DELETE: Method.DELETE,
    }
    endpoints: Endpoints = {}
    fields: Fields = {}
    serializer_class: BaseSerializer
    url_builder_class: BaseURLBuilder

    @classmethod
    @abstractmethod
    def get_serializer(cls, **kwargs) -> type[BaseSerializer]: ...

    @classmethod
    @abstractmethod
    def get_method(cls, endpoint: Endpoint, **kwargs) -> Method: ...

    @classmethod
    @abstractmethod
    def get_field(cls, field: Field) -> str: ...

    @classmethod
    @abstractmethod
    def get_pk(cls, obj: Schema | Mapping) -> any: ...

    @classmethod
    @abstractmethod
    async def create[T: Schema](
            cls,
            client: BaseRESTClient,
            obj: Schema | Mapping,
            response_type: ResponseType = None,
            **kwargs,
    ) -> T | None: ...

    @classmethod
    @abstractmethod
    async def read[T: Schema](
            cls,
            client: BaseRESTClient,
            response_type: ResponseType = None,
            **kwargs,
    ) -> Iterable[T]: ...

    @classmethod
    @abstractmethod
    async def read_one[T: Schema](
            cls,
            client: BaseRESTClient,
            obj_or_pk: Schema | Mapping | any,
            response_type: ResponseType = None,
            **kwargs,
    ) -> T: ...

    @classmethod
    @abstractmethod
    async def update[T: Schema](
            cls,
            client: BaseRESTClient,
            obj: Schema | Mapping,
            response_type: ResponseType = None,
            **kwargs,
    ) -> T | None: ...

    @classmethod
    @abstractmethod
    async def delete[T: Schema](
            cls,
            client: BaseRESTClient,
            obj_or_pk: Schema | Mapping | any,
            response_type: ResponseType = None,
            **kwargs,
    ) -> T | None: ...



from resty.types import Response


class RestyError(Exception):
    pass


class NetworkError(RestyError):
    pass


class URLBuildingError(RestyError):
    pass


class ConnectError(NetworkError, ConnectionError):
    def __init__(self, url: str):
        self.url = url  # pragma: no cover
        super().__init__(
            f"Failed to establish a connection to the server {url}"
        )  # pragma: no cover


class HTTPError(NetworkError):
    def __init__(self, response: Response):
        self.response = response  # pragma: no cover
        super().__init__(  # pragma: no cover
            f"{response.request.method.value}: {response.request.url} -> {response.status}"
        )


class BadRequestError(HTTPError):
    pass


class UnauthorizedError(HTTPError):
    pass


class ForbiddenError(HTTPError):
    pass


class NotFoundError(HTTPError):
    pass


class MethodNotAllowedError(HTTPError):
    pass


class InternalServerError(HTTPError):
    pass



from resty.managers.types import BaseManager, BaseURLBuilder
from resty.managers.managers import Manager
from resty.managers.builders import URLBuilder

__all__ = [
    "BaseManager",
    "Manager",
    "BaseURLBuilder",
    "URLBuilder",
]



from typing import Iterable, Mapping

from resty.serializers.types import BaseSerializer
from resty.types import Schema


class Serializer(BaseSerializer):
    @classmethod
    def serialize(cls, obj: Schema, **kwargs) -> Mapping:
        return obj.model_dump()

    @classmethod
    def serialize_many(cls, objs: Iterable[Schema], **kwargs) -> Iterable:
        return tuple(cls.serialize(obj=obj, **kwargs) for obj in objs)

    @classmethod
    def deserialize[T: Schema](cls, schema: type[T], data: Mapping, **kwargs) -> T:
        return schema.model_validate(data)

    @classmethod
    def deserialize_many[
        T: Schema
    ](cls, schema: type[T], data: Iterable, **kwargs,) -> Iterable[T]:
        return tuple(
            cls.deserialize(schema=schema, data=dataset, **kwargs) for dataset in data
        )
